/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _MOD_SMTPD_H
#define _MOD_SMTPD_H

#include "apr_pools.h"
#include "apr_hash.h"
#include "apr_file_io.h"
#include "httpd.h"

#ifdef __cplusplus
extern "C" {
#endif

  /* Current version of the Plugin interface */
#define SMTPD_PLUGIN_VERSION 20050414

#if !defined(WIN32)
#define SMTPD_DECLARE(type)            type
#define SMTPD_DECLARE_NONSTD(type)     type
#define SMTPD_DECLARE_DATA
#elif defined(SMTPD_DECLARE_STATIC)
#define SMTPD_DECLARE(type)            type __stdcall
#define SMTPD_DECLARE_NONSTD(type)     type
#define SMTPD_DECLARE_DATA
#elif defined(SMTPD_DECLARE_EXPORT)
#define SMTPD_DECLARE(type)            __declspec(dllexport) type __stdcall
#define SMTPD_DECLARE_NONSTD(type)     __declspec(dllexport) type
#define SMTPD_DECLARE_DATA             __declspec(dllexport)
#else
#define SMTPD_DECLARE(type)            __declspec(dllimport) type __stdcall
#define SMTPD_DECLARE_NONSTD(type)     __declspec(dllimport) type
#define SMTPD_DECLARE_DATA             __declspec(dllimport)
#endif

  typedef enum {
    SMTPD_DECLINED,
    SMTPD_DENY,
    SMTPD_DENYSOFT,
    SMTPD_DENY_DISCONNECT,
    SMTPD_DENYSOFT_DISCONNECT,
    SMTPD_DONE,
    SMTPD_DONE_DISCONNECT,
    SMTPD_OK
  } smtpd_retcode;

  typedef enum {
    SMTPD_STATE_GOT_NOTHING,
    SMTPD_STATE_GOT_HELO,
    SMTPD_STATE_GOT_MAIL,
    SMTPD_STATE_GOT_RCPT
  } smtpd_trans_state;

  typedef enum {
    SMTPD_PROTOCOL_SMTP,
    SMTPD_PROTOCOL_ESMTP
  } smtpd_protocol_type;

  typedef struct smtpd_return_data {
    apr_pool_t *p;
    /* list of messages
       null terminated */
    char **msgs;
  } smtpd_return_data;

  typedef struct smtpd_trans_rec {
    apr_pool_t *p;

    /* where are we in the current transaction */
    smtpd_trans_state state_vector;

    /* is this esmtp or smtp */
    /* by default smtp */
    smtpd_protocol_type extended;

    /* hostname we were helo'd with */
    char *helo;

    /* string of who this mail is from */
    char *mail_from;

    /* recipient list */
    apr_array_header_t *rcpt_to;

    /* spooled data file pointer */
    apr_file_t *tfp;
  } smtpd_trans_rec;

  typedef struct smtpd_conn_rec {
    apr_pool_t *p;
    conn_rec *c;
    server_rec *s;
   
    /* extensions registered with this connection */
    apr_array_header_t *extensions;

    /* current transaction */
    smtpd_trans_rec *transaction;

    /* bb in */
    apr_bucket_brigade *bb_in;

    /* bb out */
    apr_bucket_brigade *bb_out;
  } smtpd_conn_rec;
  
  /* public */
  SMTPD_DECLARE_NONSTD(void)
  smtpd_register_extension(smtpd_conn_rec *scr, const char *line);
       
  SMTPD_DECLARE_NONSTD(void)
  smtpd_reset_transaction(smtpd_conn_rec *scr);

  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode,
			    unrecognized_command,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *command, char *data));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, connect,
			    (smtpd_conn_rec *scr, smtpd_return_data *in));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, reset_transaction,
			    (smtpd_conn_rec *scr));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, helo,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *str));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, ehlo,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *str));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, mail,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *str));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, rcpt,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *str));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, vrfy,
			    (smtpd_conn_rec *scr, smtpd_return_data *in,
			     char *str));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, quit,
			    (smtpd_conn_rec *scr, smtpd_return_data *in));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, data,
  			    (smtpd_conn_rec *scr, smtpd_return_data *in));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, data_post,
			    (smtpd_conn_rec *scr, smtpd_return_data *in));
  APR_DECLARE_EXTERNAL_HOOK(smtpd, SMTPD, smtpd_retcode, queue,
			    (smtpd_conn_rec *scr, smtpd_return_data *in));
  

#ifdef __cplusplus
}
#endif

#endif /*_MOD_SMTPD_H */
