/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define CORE_PRIVATE
#include "apr_strings.h"
#include "apr_pools.h"
#include "httpd.h"
#include "http_protocol.h"
#include "http_config.h"
#include "http_connection.h"
#include "http_core.h"
#include "http_request.h"
#include "http_log.h"
#include "ap_config.h"
#include "ap_mmn.h"
#include "apr_lib.h"
#include "apr_buckets.h"
#include "apr_errno.h"
#include "scoreboard.h"

#include "smtp.h"
#include "mod_smtpd.h"

extern module AP_MODULE_DECLARE_DATA smtpd_module;

inline static int
smtpd_handle_unrecognized_command(smtpd_conn_rec *scr,
				  smtpd_return_data *in_data,
				  char *command, char *data);

#define BUFFER_STR_LEN 1024
void
smtpd_process_connection_internal(smtpd_conn_rec *scr)
{
  apr_pool_t *p;
  apr_hash_t *handlers = smtpd_get_handlers();
  char cmdbuff[BUFFER_STR_LEN];
  smtpd_handler_st *handle_func;
  char *buffer = cmdbuff;
  char *command;
  smtpd_return_data in_data;
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(scr->s->module_config,
			 &smtpd_module);

  apr_pool_create(&p, scr->p);
  in_data.p = p;

  in_data.msgs = NULL;
  switch(smtpd_run_connect(scr, &in_data)) {
  case SMTPD_DENY:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "Connection Denied");
    if (in_data.msgs) {
      smtpd_respond_multiline(scr, 550, in_data.msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "Connection from you denied, bye bye.");
    }
    goto end;
  case SMTPD_DENYSOFT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "Connection Denied");
    if (in_data.msgs) {
      smtpd_respond_multiline(scr, 450, in_data.msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "Connection from you temporarily "
			    "denied, bye bye.");
    }
    goto end;
  case SMTPD_DONE:
    break;
  case SMTPD_DONE_DISCONNECT:
    goto end;
  default:
    sprintf(buffer, "%s %s", scr->s->server_hostname, pConfig->sId);
    smtpd_respond_oneline(scr, 220, buffer);
    break;
  }

  while (smtpd_getline(scr, buffer, BUFFER_STR_LEN) == APR_SUCCESS) {
    apr_pool_clear(p);
    command = ap_getword_white_nc(p, &buffer);
    ap_str_tolower(command);
    handle_func = apr_hash_get(handlers, command, APR_HASH_KEY_STRING);

    in_data.msgs = NULL;
    /* command not recognized */
    if (!handle_func)  {
      if (!smtpd_handle_unrecognized_command(scr, &in_data, command, buffer))
	break;
    } else {
      if (!handle_func->func(scr, buffer, &in_data, handle_func->data))
	break;
    }
    
    buffer = cmdbuff;
  }

 end:
  return;
}

inline static int
smtpd_handle_unrecognized_command(smtpd_conn_rec *scr,
				  smtpd_return_data *in_data,
				  char *command, char *data) {
  switch(smtpd_run_unrecognized_command(scr, in_data, command, data)) {
  case SMTPD_DENY:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 521, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 521, "Command Denied");
    }
    return 521;
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  default:
    smtpd_respond_oneline(scr, 500, "Syntax error, command unrecognized");
    return 500;
  }

}

HANDLER_DECLARE(ehlo) {
  int i = 0;
  smtpd_trans_rec *str = scr->transaction;

  if (buffer[0] == '\0') {
    smtpd_respond_oneline(scr, 501, "Syntax: EHLO hostname");
    return 501;
  }
  
  switch(smtpd_run_ehlo(scr, in_data, buffer)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  case SMTPD_DENY:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "");
    }
    return 550;
  case SMTPD_DENYSOFT:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "");
    }
    return 450;
  default:
    break;
  }

  /* default behavior: */

  /* RFC 2821 states that when ehlo or helo is received, reset */
  /* state */
  smtpd_reset_transaction(scr);

  str->helo = apr_pstrdup(str->p, buffer);
  str->state_vector = SMTPD_STATE_GOT_HELO;
  str->extended = SMTPD_PROTOCOL_ESMTP;

  if (scr->extensions->nelts) {
    char **messages = apr_palloc(in_data->p, sizeof(char *) *
				 (scr->extensions->nelts + 2));
    
    messages[0] = str->helo;
    for (i = 0; i < scr->extensions->nelts; ++i) {
      messages[i+1] = ((char **)scr->extensions->elts)[i];
    }
    messages[i] = 0;

    smtpd_respond_multiline(scr, 250, messages);
  } else {
    smtpd_respond_oneline(scr, 250, str->helo);
  }

  return 250;
}

HANDLER_DECLARE(helo) {
  smtpd_trans_rec *str = scr->transaction;

  /* bad syntax */
  if (buffer[0] == '\0') {
    smtpd_respond_oneline(scr, 501, "Syntax: HELO hostname");
    return 501;
  }

  switch(smtpd_run_helo(scr, in_data, buffer)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  case SMTPD_DENY:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "");
    }
    return 550;
  case SMTPD_DENYSOFT:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "");
    }
    return 450;
  default:
    break;
  }

  /* RFC 2821 states that when ehlo or helo is received, reset */
  /* state */
  smtpd_reset_transaction(scr);
  
  str->helo = apr_pstrdup(str->p, buffer);
  str->state_vector = SMTPD_STATE_GOT_HELO;
  smtpd_respond_oneline(scr, 250, str->helo);

  return 250;
}

HANDLER_DECLARE(mail) {
  smtpd_trans_rec *str = scr->transaction;
  char *loc;

  /* already got mail */
  if (str->state_vector == SMTPD_STATE_GOT_MAIL) {
    smtpd_respond_oneline(scr, 503, "Error: Nested MAIL command");
    return 503;
  }

  /* bad syntax */
  if ((loc = ap_strcasestr(buffer, "from:")) == NULL) {
    smtpd_respond_oneline(scr, 501,  "Syntax: MAIL FROM:<address>");
    return 501;
  }

  loc += sizeof("from:") - 1;

  ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
	       "full from_parameter: %s", loc);  

  switch(smtpd_run_mail(scr, in_data, loc)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    /* zero to disconnect */
    return 0;
  case SMTPD_DENY:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "deny mail from %s (%s)", loc,
		 in_data->msgs ? in_data->msgs[0] : "");  
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "denied");
    }
    return 550;
  case SMTPD_DENYSOFT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "denysoft mail from %s (%s)", loc,
		 in_data->msgs ? in_data->msgs[0] : "");  
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "temporarily denied");
    }
    return 450;
  case SMTPD_DENY_DISCONNECT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "deny mail from %s (%s)", loc,
		 in_data->msgs ? in_data->msgs[0] : "");  
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "denied");
    }
    /* zero to disconnect */
    return 0;
  case SMTPD_DENYSOFT_DISCONNECT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "denysoft mail from %s (%s)", loc,
		 in_data->msgs[0] ? in_data->msgs[0] : "");  
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "temporarily denied");
    }
    /* zero to disconnect */
    return 0;
  default:
    break;
  }

  /* default handling */
  str->mail_from = apr_pstrdup(str->p, loc);
  str->state_vector = SMTPD_STATE_GOT_MAIL;
  smtpd_respond_oneline(scr, 250, "Ok");

  return 250;
}

HANDLER_DECLARE(rcpt) {
  smtpd_trans_rec *str = scr->transaction;
  char *loc;

  /* need mail first */
  if ((str->state_vector != SMTPD_STATE_GOT_MAIL) &&
      (str->state_vector != SMTPD_STATE_GOT_RCPT)) {
    smtpd_respond_oneline(scr, 503, "Error: need MAIL command");
    return 503;
  }

  /* bad syntax */
  if ((loc = ap_strcasestr(buffer, "to:")) == NULL) {
    smtpd_respond_oneline(scr, 501, "Syntax: RCPT TO:<address>");
    return 501;
  }

  loc += sizeof("to:") - 1;

  switch(smtpd_run_rcpt(scr, in_data, loc)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  case SMTPD_DENY:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "relaying denied");
    }
    return 550;
  case SMTPD_DENYSOFT:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "relaying temporarily denied");
    }
    return 450;
  case SMTPD_DENY_DISCONNECT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "delivery denied (%s)",
		 in_data->msgs ? in_data->msgs[0] : "");
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 550, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 550, "relaying denied");
    }
    return 0;
  case SMTPD_DENYSOFT_DISCONNECT:
    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, scr->s,
		 "delivery denied (%s)",
		 in_data->msgs ? in_data->msgs[0] : "");
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 450, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 450, "relaying temporarily denied");
    }
    return 0;
  case SMTPD_OK: /* recipient is okay */
    break;
  default:
    smtpd_respond_oneline(scr, 450,
			  "No plugin decided if relaying is allowed");
    return 450;
  }

  /* add a recipient */
  (*((char **)apr_array_push(str->rcpt_to))) = apr_pstrdup(str->p, loc);
  str->state_vector = SMTPD_STATE_GOT_RCPT;
  smtpd_respond_oneline(scr, 250, "Ok");

  return 250;
}


inline static int
smtpd_queue(smtpd_conn_rec *scr, smtpd_return_data *in_data) {
    switch(smtpd_run_queue(scr, in_data)) {
    case SMTPD_DONE:
      return 1;
    case SMTPD_DONE_DISCONNECT:
      return 0;
    case SMTPD_OK:
      if (in_data->msgs) {
	smtpd_respond_multiline(scr, 250, in_data->msgs);
      } else {
	smtpd_respond_oneline(scr, 250, "Queued");
      }
      return 250;
    case SMTPD_DENY:
      if (in_data->msgs) {
	smtpd_respond_multiline(scr, 552, in_data->msgs);
      } else {
	smtpd_respond_oneline(scr, 552, "Message denied");
      }
      return 552;
    case SMTPD_DENYSOFT:
      if (in_data->msgs) {
	smtpd_respond_multiline(scr, 452, in_data->msgs);
      } else {
	smtpd_respond_oneline(scr, 452, "Message denied temporarily");
      }
      return 452;
    default:
      if (in_data->msgs) {
	smtpd_respond_multiline(scr, 451, in_data->msgs);
      } else {
	smtpd_respond_oneline(scr, 451,
			      "Queuing declined or disabled; try again later");
      }
      return 451;
    }
}


HANDLER_DECLARE(data) {
  smtpd_trans_rec *str = scr->transaction;
  smtpd_svr_config_rec *pConfig =
    ap_get_module_config(scr->s->module_config,
			 &smtpd_module);
  int rv, retval = 0;
  char *tempfile;
  apr_bucket_brigade *bb;
  apr_file_t *tfp;
  apr_size_t len, total_data = 0;

  switch(smtpd_run_data(scr, in_data)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  case SMTPD_DENY:
    /* REVIEW: should we reset state here? */
    /* smtpd_clear_request_rec(sr); */
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 554, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 554, "Message denied");
    }
    return 554;
  case SMTPD_DENYSOFT:
    /* REVIEW: should we reset state here? */
    /* smtpd_clear_request_rec(sr); */
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 451, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 451, "Message denied temporarily");
    }
    return 451;
  case SMTPD_DENY_DISCONNECT:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 554, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 554, "Message denied");
    }
    return 0;
  case SMTPD_DENYSOFT_DISCONNECT:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 451, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 451, "Message denied temporarily");
    }
    return 0;
  default:
    break;
  }

  if (str->state_vector != SMTPD_STATE_GOT_RCPT) {
    smtpd_respond_oneline(scr, 503, "Error: need RCPT command");
    return 503;
  }

  smtpd_respond_oneline(scr, 354, "End data with <CR><LF>.<CR><LF>");
  
  bb = apr_brigade_create(scr->p, scr->c->bucket_alloc);

  tempfile = apr_pstrdup(str->p, "/tmp/tmp.XXXXXX");
  rv = apr_file_mktemp(&tfp, tempfile,
		       APR_CREATE | APR_WRITE | APR_READ |
		       APR_DELONCLOSE, scr->p);

  if (rv != APR_SUCCESS) {
    smtpd_respond_oneline(scr, 421, "Error: Internal");
    /* file error close connection */
    return 0;
  }

  /* just wait until we get the line with a dot. */
  /* or until we can't read anymore. */
  /* or until we have too much data */
  while ((smtpd_getline(scr, buffer, BUFFER_STR_LEN) == APR_SUCCESS) &&
	 (strcmp(buffer, ".")) &&
	 (total_data < pConfig->max_data)) {
    apr_file_write(tfp, buffer, &len);
    len = sizeof("\n");
    apr_file_write(tfp, "\n", &len);
    total_data += len;
  }

  str->tfp = tfp;
  
  switch(smtpd_run_data_post(scr, in_data)) {
  case SMTPD_DONE:
    retval = 1;
    goto cleanup;
  case SMTPD_DONE_DISCONNECT:
    retval = 0;
    goto cleanup;
  case SMTPD_DENY:
    retval = 552;
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 552, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 552, "Message denied");
    }
    break;
  case SMTPD_DENYSOFT:
    retval = 452;
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 452, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 452, "Message denied temporarily");
    }
    break;
  default:
    retval = smtpd_queue(scr, in_data);
  }

  smtpd_reset_transaction(scr);

 cleanup:
  apr_file_close(tfp);
  return retval;
}

HANDLER_DECLARE(rset) {
  smtpd_reset_transaction(scr);

  smtpd_respond_oneline(scr, 250, "Ok");

  return 250;
}

HANDLER_DECLARE(noop) {
  smtpd_respond_oneline(scr, 250, "Ok");

  return 250;
}

HANDLER_DECLARE(quit) {
  if (smtpd_run_quit(scr, in_data) != SMTPD_DONE) {
    smtpd_respond_oneline(scr, 221, "Bye");
  }
  /* zero to disconnect */
  return 0;
}

HANDLER_DECLARE(vrfy) {
 
  switch(smtpd_run_vrfy(scr, in_data, buffer)) {
  case SMTPD_DONE:
    return 1;
  case SMTPD_DONE_DISCONNECT:
    return 0;
  case SMTPD_DENY:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 554, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 554, "Address denied");
    }
    return 554;
  case SMTPD_OK: /* user is okay */
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 250, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 250, "Address okay");
    }
    return 250;
  default:
    if (in_data->msgs) {
      smtpd_respond_multiline(scr, 252, in_data->msgs);
    } else {
      smtpd_respond_oneline(scr, 252,
			    "Address seems fine, but we might not accept it.");
    }
    return 252;
  }
}
